#ifndef _FBVNCSERVER_H
#define _FBVNCSERVER_H

/* *******************************************************
 This file is part of fbvncserver.

 fbvncserver is free software; you can redistribute it
 and/or modify it under the terms of the
 GNU General Public License as published by
 the Free Software Foundation; either version 2 of the
 License, or (at your option) any later version.

 fbvncserver is distributed in the hope that it will be
 useful, but WITHOUT ANY WARRANTY; without even the
 implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.

 You should have received a copy of the GNU General Public
 License along with fbvncserver; if not, write to the
 Free Software Foundation, Inc., 51 Franklin St,
 Fifth Floor, Boston, MA  02110-1301  USA
 ********************************************************/

/**
 * @file  fbvncserver.h
 * @brief fbvncserver - global definitions
 * @author Patrik Gfeller (pgf) - CodeOmega (TM)
 */

/*
 * Standard ANSI C
 * Documentation @ http://www.cplusplus.com/ref/
 */
#include <stdlib.h>   /* general purpose (memory allocation, process control, conversion ...) */
#include <stdio.h>    /* library for Input/Output with streams                                */
#include <unistd.h>   /* standard symbolic constants and types                                */
#include <string.h>   /* library to manipulate C strings                                      */
#include <math.h>     /* library for mathematic operations (linker: -lm)                      */

#include <sys/mman.h> /* memory management declarations */
#include <sys/wait.h> /* Wait for Process Termination   */

/*
 * LibVNCServer
 * URL: http://sourceforge.net/projects/libvncserver/
 */
#include "rfb/rfb.h"
#include "rfb/keysym.h"

/* Project */
#include "config.h"      /** @file config.h build configuration     */
#include "rotate.h"      /* utility functions to rotate framebuffer */
#include "dev_uinput.h"  /* uinput kernel interface                 */

/* define constants */
#define DESKTOP_NAME "PDA" 
#ifndef TRUE
#define TRUE  1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define UINPUT_KBD_NAME "fbvnc key-injector"  /**< uinput device name for the remote keyboard */
#define UINPUT_TS_NAME "fbvnc mouse-injector" /**< uinput device name for the remote touchscreen */
#define UINPUT_INPUTAPI_NAME "fbvnc TS"       /**< uinput device name for the merged TS events (if inputAPI is used) */

#define CONFIGBASEDIR "/usr/share/vnc/"       /**< where we will find our configuration files (e.g. keymaps) */
#define DEFAULT_KBD_LAYOUT "en-us"            /**< name of the default keymap used to translate key events */

/* define MACROS */
#ifdef DEBUG
#define TRACE(STR) if (debug) {rfbLog(STR); }
#define TRACE1(STR,ARG1) if (debug) {rfbLog(STR, ARG1); }
#define TRACE2(STR,ARG1,ARG2) if (debug) {rfbLog(STR, ARG1, ARG2); }
#define TRACE3(STR,ARG1,ARG2,ARG3) if (debug) {rfbLog(STR, ARG1, ARG2, ARG3); }
#define TRACE4(STR,ARG1,ARG2,ARG3,ARG4) if (debug) {rfbLog(STR, ARG1, ARG2, ARG3,ARG4); }
#define TRACE5(STR,ARG1,ARG2,ARG3,ARG4, ARG5) if (debug) {rfbLog(STR, ARG1, ARG2, ARG3,ARG4, ARG5); }
#else
#define TRACE(STR)
#define TRACE1(STR,ARG1)
#define TRACE2(STR,ARG1,ARG2)
#define TRACE3(STR,ARG1,ARG2, ARG3)
#define TRACE4(STR,ARG1,ARG2, ARG3, ARG4)
#define TRACE5(STR,ARG1,ARG2, ARG3, ARG4, ARG5)
#endif

 static unsigned short int rotate = 0; /**< rotate screen x-times 90° cw    */
 static unsigned short int debug  = 0; /**< display debug information [0,1] */
 static unsigned short int uinput = 1; /**< try to use uinput [0,1]         */

 static int uinput_fd_kbd   = -1; /**< fd to inject keyboard events into the kernel (uinput) */
 static int uinput_fd_mouse = -1; /**< fd to inject mouse events into the kernel (uinput)    */

/**
 @dir src SourceCode
 */

/**
 @mainpage PDA fbvncserver Project
 @author 2003-2007 &copy; by Patrik Gfeller (pgf), CodeOmega (tm)

<hr >

<h1> Structure </h1>
<h2> Keyboard input </h2>

With showkey you can see what keycodes are produced on the console.
The keycodes are the result of a translation of the scancodes,
which are the actual output of the keyboard.
The scancodes can be made visible by showkey -s, and by getkeycodes.

The system keymap is the translation table which converts keycodes to strings.
The system keymap is a plain ASCII file which is loaded by loadkeys
(use dumpkeys to show it).

<pre>

                 +--------+
[TouchScreen] => | Kernel | ---> [tslib] ---> [tssimd.h/.c] ---> [tslib]     =>
                 +--------+                        A
                                                   |
                                                   |
           ptrevent()                              |
 VNC => [fbvncserver.h/.c] ------------------------+
           keyevent()                              |
                                                   | keysym2scancode()
                                                   |
                                                   V
                                          |---- uinput ----|
                                          |---- event -----|
                                          |-- kbd driver --| (translates the scancodes again)
|---kbd/bios/firmware---|-----------------+-----kernel-----+---------------| =>

</pre>

<pre>
 #:> dumpkeys --long-info
 #:> getkeycodes
</pre>
<hr >
 Based on the work of Pierre-Philippe Coupard, who as written the
 original iPAQ36xx &amp; Zaurus code; fragments used from <b>x11vnc</b>,
 <b>kbdd</b>, <b>tslib</b>, ... <br >
 The core of this SW is <b><a href="http://sourceforge.net/projects/libvncserver/"> LibVNCServer </a></b>.

 additional thanks to
  - <a href="http://maxima.sourceforge.net"> MAXIMA CAS </a> for resolving the calibration equations :-)

 <hr >

 <h2> Requirements &amp; dependencies: </h2>
 <h3> Build </h3>
   - libvncserver
   - zlib, Compression Library (http://www.zlib.net/)
   - tslib
   - gmp, GNU Multiple Precision Arithmetic Library

./configure --host=i686-linux

 <h3> Runtime</h3>
  - libgmp
  - tslib
  - zlib

 The core of this SW is <b><a href="http://sourceforge.net/projects/libvncserver/"> LibVNCServer </a></b>.

 @todo
  -  [00%] tssimd.sh - react smart, if e.g. the uinput subsystem is not available ...
  -  [00%] replace ts configuration (& restore the org. if the package is removed)
  -  [85%] use uinput to replace keyboard Kernel module
  -  [85%] support for different keymaps
  -  [00%] create/use human readable config file
  -  [85%] use GNU build system
  -  [40%] create open embedded .bb
  -  [00%] copy/paste support
  -  [00%] text chat
<h3> already done </h3>
  - [100%] use /etc/pointercal for ts calibration (linear), 15.06.2007
 */
#endif
